@@ -11,20 +11,18 @@ const Meta = imports.gi.Meta;
 const Pango = imports.gi.Pango;
 const Signals = imports.signals;
 const Shell = imports.gi.Shell;
+const St = imports.gi.St;
 
-const Button = imports.ui.button;
 const DND = imports.ui.dnd;
-const Link = imports.ui.link;
 const Main = imports.ui.main;
 
-const ITEM_DISPLAY_NAME_COLOR = new Clutter.Color();
-ITEM_DISPLAY_NAME_COLOR.from_pixel(0xffffffff);
+const RedisplayFlags = { NONE: 0,
+                         FULL: 1 << 1,
+                         SUBSEARCH: 1 << 2,
+                         IMMEDIATE: 1 << 3 };
+
 const ITEM_DISPLAY_DESCRIPTION_COLOR = new Clutter.Color();
 ITEM_DISPLAY_DESCRIPTION_COLOR.from_pixel(0xffffffbb);
-const ITEM_DISPLAY_BACKGROUND_COLOR = new Clutter.Color();
-ITEM_DISPLAY_BACKGROUND_COLOR.from_pixel(0x00000000);
-const ITEM_DISPLAY_SELECTED_BACKGROUND_COLOR = new Clutter.Color();
-ITEM_DISPLAY_SELECTED_BACKGROUND_COLOR.from_pixel(0x4f6fadaa);
 const DISPLAY_CONTROL_SELECTED_COLOR = new Clutter.Color();
 DISPLAY_CONTROL_SELECTED_COLOR.from_pixel(0x112288ff);
 const PREVIEW_BOX_BACKGROUND_COLOR = new Clutter.Color();
@@ -32,10 +30,7 @@ PREVIEW_BOX_BACKGROUND_COLOR.from_pixel(0xADADADf0);
 
 const DEFAULT_PADDING = 4;
 
-const ITEM_DISPLAY_HEIGHT = 50;
 const ITEM_DISPLAY_ICON_SIZE = 48;
-const ITEM_DISPLAY_PADDING = 1;
-const ITEM_DISPLAY_PADDING_RIGHT = 2;
 const DEFAULT_COLUMN_GAP = 6;
 
 const PREVIEW_ICON_SIZE = 96;
@@ -46,8 +41,6 @@ const PREVIEW_BOX_CORNER_RADIUS = 10;
 const PREVIEW_PLACING = 3/4;
 const PREVIEW_DETAILS_MIN_WIDTH = PREVIEW_ICON_SIZE * 2;
 
-const INFORMATION_BUTTON_SIZE = 16;
-
 /* This is a virtual class that represents a single display item containing
  * a name, a description, and an icon. It allows selecting an item and represents
  * it by highlighting it with a different background color than the default.
@@ -58,12 +51,8 @@ function GenericDisplayItem() {
 
 GenericDisplayItem.prototype = {
     _init: function() {
-        this.actor = new Big.Box({ orientation: Big.BoxOrientation.HORIZONTAL,
-                                   spacing: ITEM_DISPLAY_PADDING,
-                                   reactive: true,
-                                   background_color: ITEM_DISPLAY_BACKGROUND_COLOR,
-                                   corner_radius: 4,
-                                   height: ITEM_DISPLAY_HEIGHT });
+        this.actor = new St.BoxLayout({ style_class: "generic-display-item",
+                                         reactive: true });
 
         this.actor._delegate = this;
         this.actor.connect('button-release-event',
@@ -75,53 +64,20 @@ GenericDisplayItem.prototype = {
                                      }));
 
         let draggable = DND.makeDraggable(this.actor);
-        draggable.connect('drag-begin', Lang.bind(this, this._onDragBegin));
 
-        this._infoContent = new Big.Box({ orientation: Big.BoxOrientation.HORIZONTAL,
-                                          spacing: DEFAULT_PADDING });
-        this.actor.append(this._infoContent, Big.BoxPackFlags.EXPAND);
+        this._iconBin = new St.Bin();
+        this.actor.add(this._iconBin);
 
-        this._iconBox = new Big.Box();
-        this._infoContent.append(this._iconBox, Big.BoxPackFlags.NONE);
-
-        this._infoText = new Big.Box({ orientation: Big.BoxOrientation.VERTICAL,
-                                       spacing: DEFAULT_PADDING });
-        this._infoContent.append(this._infoText, Big.BoxPackFlags.EXPAND);
-
-        let global = Shell.Global.get();
-        let infoIconUri = "file://" + global.imagedir + "info.svg";
-        let infoIcon = Shell.TextureCache.get_default().load_uri_sync(Shell.TextureCachePolicy.FOREVER,
-                                                                      infoIconUri,
-                                                                      INFORMATION_BUTTON_SIZE,
-                                                                      INFORMATION_BUTTON_SIZE);
-        this._informationButton = new Button.iconButton(this.actor, INFORMATION_BUTTON_SIZE, infoIcon);
-        let buttonBox = new Big.Box({ width: INFORMATION_BUTTON_SIZE + 2 * DEFAULT_PADDING,
-                                      height: INFORMATION_BUTTON_SIZE,
-                                      padding_left: DEFAULT_PADDING, padding_right: DEFAULT_PADDING,
-                                      y_align: Big.BoxAlignment.CENTER });
-        buttonBox.append(this._informationButton.actor, Big.BoxPackFlags.NONE);
-        this.actor.append(buttonBox, Big.BoxPackFlags.END);
-
-        // Connecting to the button-press-event for the information button ensures that the actor,
-        // which is a draggable actor, does not get the button-press-event and doesn't initiate
-        // the dragging, which then prevents us from getting the button-release-event for the button.
-        this._informationButton.actor.connect('button-press-event',
-                                              Lang.bind(this,
-                                                        function() {
-                                                            return true;
-                                                        }));
-        this._informationButton.actor.connect('button-release-event',
-                                              Lang.bind(this,
-                                                        function() {
-                                                            // Selects the item by highlighting it and displaying its details
-                                                            this.emit('show-details');
-                                                            return true;
-                                                        }));
+        this._infoText = new St.BoxLayout({ style_class: 'generic-display-item-text',
+                                             vertical: true });
+        this.actor.add(this._infoText, { expand: true, y_fill: false });
 
         this._name = null;
         this._description = null;
         this._icon = null;
 
+        this._initialLoadComplete = false;
+
         // An array of details description actors that we create over time for the item.
         // It is used for updating the description text inside the details actor when
         // the description text for the item is updated.
@@ -146,24 +102,10 @@ GenericDisplayItem.prototype = {
 
     //// Public methods ////
 
-    // Shows the information button when the item was drawn under the mouse pointer.
-    onDrawnUnderPointer: function() {
-        this._informationButton.show();
-    },
-
     // Highlights the item by setting a different background color than the default 
     // if isSelected is true, removes the highlighting otherwise.
     markSelected: function(isSelected) {
-       let color;
-       if (isSelected) {
-           color = ITEM_DISPLAY_SELECTED_BACKGROUND_COLOR;
-           this._informationButton.forceShow(true);
-       }
-       else {
-           color = ITEM_DISPLAY_BACKGROUND_COLOR;
-           this._informationButton.forceShow(false);
-       }
-       this.actor.background_color = color;
+        this.actor.set_style_pseudo_class(isSelected ? "selected" : null);
     },
 
     /*
@@ -179,19 +121,14 @@ GenericDisplayItem.prototype = {
                                         spacing: PREVIEW_BOX_SPACING });
 
         // Inner box with name and description
-        let textDetails = new Big.Box({ orientation: Big.BoxOrientation.VERTICAL,
-                                        spacing: PREVIEW_BOX_SPACING });
-        let detailsName = new Clutter.Text({ color: ITEM_DISPLAY_NAME_COLOR,
-                                             font_name: "Sans bold 14px",
-                                             line_wrap: true,
-                                             text: this._name.text });
-        textDetails.append(detailsName, Big.BoxPackFlags.NONE);
+        let textDetails = new St.BoxLayout({ style_class: 'generic-display-details',
+                                             vertical: true });
+        let detailsName = new St.Label({ style_class: 'generic-display-details-name',
+                                          text: this._name.text });
+        textDetails.add(detailsName);
 
-        let detailsDescription = new Clutter.Text({ color: ITEM_DISPLAY_NAME_COLOR,
-                                                    font_name: "Sans 14px",
-                                                    line_wrap: true,
-                                                    text: this._description.text });
-        textDetails.append(detailsDescription, Big.BoxPackFlags.NONE);
+        let detailsDescription = new St.Label({ text: this._description.text });
+        textDetails.add(detailsDescription);
 
         this._detailsDescriptions.push(detailsDescription);
 
@@ -217,7 +154,7 @@ GenericDisplayItem.prototype = {
 
     // Destroys the item.
     destroy: function() {
-      this.actor.destroy();
+        this.actor.destroy();
     },
 
     //// Pure virtual public methods ////
@@ -254,20 +191,15 @@ GenericDisplayItem.prototype = {
         }
 
         this._icon = this._createIcon();
-        this._iconBox.append(this._icon, Big.BoxPackFlags.NONE);
+        this._iconBin.set_child(this._icon);
 
-        this._name = new Clutter.Text({ color: ITEM_DISPLAY_NAME_COLOR,
-                                        font_name: "Sans 14px",
-                                        ellipsize: Pango.EllipsizeMode.END,
-                                        text: nameText });
-        this._infoText.append(this._name, Big.BoxPackFlags.EXPAND);
+        this._name = new St.Label({ style_class: "generic-display-item-name",
+                                     text: nameText });
+        this._infoText.add(this._name);
 
-        this._description = new Clutter.Text({ color: ITEM_DISPLAY_DESCRIPTION_COLOR,
-                                               font_name: "Sans 12px",
-                                               ellipsize: Pango.EllipsizeMode.END,
-                                               text: descriptionText ? descriptionText : ""
-                                            });
-        this._infoText.append(this._description, Big.BoxPackFlags.EXPAND);
+        this._description = new St.Label({ style_class: "generic-display-item-description",
+                                            text: descriptionText ? descriptionText : "" });
+        this._infoText.add(this._description);
     },
 
     // Sets the description text for the item, including the description text
@@ -299,67 +231,76 @@ GenericDisplayItem.prototype = {
     // Returns a preview icon for the item.
     _createPreviewIcon: function() {
         throw new Error("Not implemented");
-    },
+    }
 
     //// Private methods ////
-
-    // Hides the information button once the item starts being dragged.
-    _onDragBegin : function (draggable, time) {
-        // For some reason, we are not getting leave-event signal when we are dragging an item,
-        // so we should remove the link manually.
-        this._informationButton.actor.hide();
-    } 
 };
 
 Signals.addSignalMethods(GenericDisplayItem.prototype);
 
+const GenericDisplayFlags = {
+    DISABLE_VSCROLLING: 1 << 0
+}
+
 /* This is a virtual class that represents a display containing a collection of items
  * that can be filtered with a search string.
  */
-function GenericDisplay() {
-    this._init();
+function GenericDisplay(flags) {
+    this._init(flags);
 }
 
 GenericDisplay.prototype = {
-    _init : function() {
+    _init : function(flags) {
+        let disableVScrolling = (flags & GenericDisplayFlags.DISABLE_VSCROLLING) != 0;
         this._search = '';
         this._expanded = false;
 
-        this._maxItemsPerPage = null;
-        this._list = new Shell.OverflowList({ spacing: 6.0,
-                                              item_height: ITEM_DISPLAY_HEIGHT });
+        if (disableVScrolling) {
+            this.actor = this._list = new Shell.OverflowList({ spacing: 6,
+                                                                 item_height: 50 });
+        } else {
+            this.actor = new St.ScrollView({ x_fill: true, y_fill: true });
+            this.actor.get_hscroll_bar().hide();
+            this._list = new St.BoxLayout({ style_class: 'generic-display-container',
+                                             vertical: true });
+            this.actor.add_actor(this._list);
+        }
 
-        this._list.connect('notify::n-pages', Lang.bind(this, function () {
-            this._updateDisplayControl(true);
-        }));
-        this._list.connect('notify::page', Lang.bind(this, function () {
-            this._updateDisplayControl(false);
-        }));
+        this._pendingRedisplay = RedisplayFlags.NONE;
+        this.actor.connect('notify::mapped', Lang.bind(this, this._onMappedNotify));
 
         // map<itemId, Object> where Object represents the item info
         this._allItems = {};
-        // an array of itemIds of items that match the current request
-        // in the order in which the items should be displayed
-        this._matchedItems = [];
+        // set<itemId>
+        this._matchedItems = {};
+        // sorted array of items matched by search
+        this._matchedItemKeys = [];
         // map<itemId, GenericDisplayItem>
         this._displayedItems = {};
         this._openDetailIndex = -1;
         this._selectedIndex = -1;
-        // These two are public - .actor is the normal "actor subclass" property,
-        // but we also expose a .displayControl actor which is separate.
-        // See also getNavigationArea.
-        this.actor = this._list;
-        this.displayControl = new Big.Box({ background_color: ITEM_DISPLAY_BACKGROUND_COLOR,
-                                            spacing: 12,
-                                            orientation: Big.BoxOrientation.HORIZONTAL});
     },
 
     //// Public methods ////
 
     // Sets the search string and displays the matching items.
     setSearch: function(text) {
-        this._search = text.toLowerCase();
-        this._redisplay(true);
+        let lowertext = text.toLowerCase();
+        if (lowertext == this._search) {
+            return;
+        }
+        let flags = RedisplayFlags.IMMEDIATE;
+        if (this._search != '') {
+            // Because we combine search terms with OR, we have to be sure that no new term
+            // was introduced before deciding that the new search results will be a subset of
+            // the existing search results.
+            if (lowertext.indexOf(this._search) == 0 &&
+                lowertext.split(/\s+/).length == this._search.split(/\s+/).length) {
+                flags |= RedisplayFlags.SUBSEARCH;
+            }
+        }
+        this._search = lowertext;
+        this._redisplay(flags);
     },
 
     // Launches the item that is currently selected, closing the Overview
@@ -376,7 +317,7 @@ GenericDisplay.prototype = {
     // to the bottom one. Returns true if the selection actually moved up, false if it wrapped 
     // around to the bottom.
     selectUp: function() {
-        let count = this._list.displayedCount;
+        let count = this._getVisibleCount();
         let selectedUp = true;
         let prev = this._selectedIndex - 1;
         if (this._selectedIndex <= 0) {
@@ -391,7 +332,7 @@ GenericDisplay.prototype = {
     // to the top one. Returns true if the selection actually moved down, false if it wrapped 
     // around to the top.
     selectDown: function() {
-        let count = this._list.displayedCount;
+        let count = this._getVisibleCount();
         let selectedDown = true;
         let next = this._selectedIndex + 1;
         if (this._selectedIndex == count - 1) {
@@ -410,7 +351,7 @@ GenericDisplay.prototype = {
 
     // Selects the last item among the displayed items.
     selectLastItem: function() {
-        let count = this._list.displayedCount;
+        let count = this._getVisibleCount();
         if (this.hasItems())
             this._selectIndex(count - 1);
     },
@@ -431,22 +372,24 @@ GenericDisplay.prototype = {
         // positive number when this._mathedItems.length is 0
         // This can be triggered if a search string is entered for which there are no matches.
         // log("this._mathedItems.length: " + this._matchedItems.length + " this._list.displayedCount " + this._list.displayedCount);
-        return this._matchedItems.length > 0;
+        return this._matchedItemKeys.length > 0;
     },
 
     getMatchedItemsCount: function() {
-        return this._matchedItems.length;
+        return this._matchedItemKeys.length;
     },
 
     // Load the initial state
     load: function() {
-        this._redisplay(true);
+        this._redisplay(RedisplayFlags.FULL);
     },
 
     // Should be called when the display is closed
     resetState: function() {
         this._filterReset();
         this._openDetailIndex = -1;
+        if (!(this.actor instanceof Shell.OverflowList))
+            this.actor.get_vscroll_bar().get_adjustment().value = 0;
     },
 
     // Returns an actor which acts as a sidebar; this is used for
@@ -460,64 +403,18 @@ GenericDisplay.prototype = {
         return item.createDetailsActor();
     },
 
-    // Displays the page specified by the pageNumber argument.
-    displayPage: function(pageNumber) {
-        // Cleanup from the previous selection, but don't unset this._selectedIndex
-        if (this.hasSelected()) {
-            this._findDisplayedByIndex(this._selectedIndex).markSelected(false);
-        }
-        this._list.page = pageNumber;
-    },
-
     //// Protected methods ////
 
-    /*
-     * Displays items that match the current request and should show up on the current page.
-     * Updates the display control to reflect the matched items set and the page selected.
-     *
-     * resetDisplayControl - indicates if the display control should be re-created because 
-     *                       the results or the space allocated for them changed. If it's false,
-     *                       the existing display control is used and only the page links are
-     *                       updated to reflect the current page selection.
-     */
-    _displayMatchedItems: function(resetDisplayControl) {
-        // When generating a new list to display, we first remove all the old
-        // displayed items which will unset the selection. So we need 
-        // to keep a flag which indicates if this display had the selection.
-        let hadSelected = this.hasSelected();
-
+    _recreateDisplayItems: function() {
         this._removeAllDisplayItems();
-        for (let i = 0; i < this._matchedItems.length; i++) {
-            this._addDisplayItem(this._matchedItems[i]);
+        this._setDefaultList();
+        for (let itemId in this._allItems) {
+            this._addDisplayItem(itemId);
         }
-
-        if (hadSelected) {
-            this._selectedIndex = -1;
-            this.selectFirstItem();
-        }
-
-        // Check if the pointer is over one of the items and display the information button if it is.
-        // We want to do this between finishing our changes to the display and the point where
-        // the display is redrawn.
-        Mainloop.idle_add(Lang.bind(this,
-                                    function() {
-                                        let [child, x, y, mask] = Gdk.Screen.get_default().get_root_window().get_pointer();
-                                        let global = Shell.Global.get();
-                                        let actor = global.stage.get_actor_at_pos(Clutter.PickMode.REACTIVE,
-                                                                                  x, y);
-                                        if (actor != null) {
-                                            let item = this._findDisplayedByActor(actor);
-                                            if (item != null) {
-                                                item.onDrawnUnderPointer();
-                                            }
-                                        }
-                                        return false;
-                                    }),
-                          Meta.PRIORITY_BEFORE_REDRAW);
     },
 
-    // Creates a display item based on the information associated with itemId 
-    // and adds it to the displayed items.
+    // Creates a display item based on the information associated with itemId
+    // and adds it to the list of displayed items, but does not yet display it.
     _addDisplayItem : function(itemId) {
         if (this._displayedItems.hasOwnProperty(itemId)) {
             log("Tried adding a display item for " + itemId + ", but an item with this item id is already among displayed items.");
@@ -531,14 +428,14 @@ GenericDisplay.prototype = {
                             Lang.bind(this,
                                       function() {
                                           // update the selection
-                                          this._selectIndex(this._list.get_actor_index(displayItem.actor));
+                                          this._selectIndex(this._list.get_children().indexOf(displayItem.actor));
                                           this.activateSelected();
                                       }));
 
         displayItem.connect('show-details',
                             Lang.bind(this,
                                       function() {
-                                          let index = this._list.get_actor_index(displayItem.actor);
+                                          let index = this._list.get_children().indexOf(displayItem.actor);
                                           /* Close the details pane if already open */
                                           if (index == this._openDetailIndex) {
                                               this._openDetailIndex = -1;
@@ -548,15 +445,15 @@ GenericDisplay.prototype = {
                                               this.emit('show-details', index);
                                           }
                                       }));
-        this._list.add_actor(displayItem.actor);
         this._displayedItems[itemId] = displayItem;
     },
 
     // Removes an item identifed by the itemId from the displayed items.
     _removeDisplayItem: function(itemId) {
-        let count = this._list.displayedCount;
+        let children = this._list.get_children();
+        let count = children.length;
         let displayItem = this._displayedItems[itemId];
-        let displayItemIndex = this._list.get_actor_index(displayItem.actor);
+        let displayItemIndex = children.indexOf(displayItem.actor);
 
         if (this.hasSelected() && count == 1) {
             this.unsetSelected();
@@ -587,33 +484,120 @@ GenericDisplay.prototype = {
         this.unsetSelected();
     },
 
+    _compareSearchMatch: function(a, b) {
+        let countA = this._matchedItems[a];
+        let countB = this._matchedItems[b];
+        if (countA > countB)
+            return -1;
+        else if (countA < countB)
+            return 1;
+        else
+            return this._compareItems(a, b);
+    },
+
+    _setMatches: function(matches) {
+        this._matchedItems = matches;
+        this._matchedItemKeys = [];
+        for (let itemId in this._matchedItems) {
+            this._matchedItemKeys.push(itemId);
+        }
+        this._matchedItemKeys.sort(Lang.bind(this, this._compareSearchMatch));
+    },
+
+    /**
+     * _redisplaySubSearch:
+     * A somewhat more optimized function called when we know
+     * that we're going to be displaying a subset of the items
+     * we already had, in the same order.  In that case, we can
+     * just hide the actors that shouldn't be shown.
+     */
+    _redisplaySubSearch: function() {
+        let matches = this._getSearchMatchedItems(true);
+
+        // Just hide all from the old set,
+        // we'll show the ones we want below
+        for (let itemId in this._displayedItems) {
+            let item = this._displayedItems[itemId];
+            item.actor.hide();
+        }
+
+        this._setMatches(matches);
+
+        for (let itemId in matches) {
+            let item = this._displayedItems[itemId];
+            item.actor.show();
+        }
+        this._list.queue_relayout();
+    },
+
+    _redisplayReordering: function() {
+        if (!this._filterActive()) {
+            this._setDefaultList();
+        } else {
+            this._setMatches(this._getSearchMatchedItems(false));
+        }
+        this._list.remove_all();
+        for (let i = 0; i < this._matchedItemKeys.length; i++) {
+            let itemId = this._matchedItemKeys[i];
+            let item = this._displayedItems[itemId];
+            item.actor.show();
+            this._list.add_actor(item.actor);
+       }
+    },
+
     /*
      * Updates the displayed items, applying the search string if one exists.
-     *
-     * resetPage - indicates if the page selection should be reset when displaying the matching results.
-     *             We reset the page selection when the change in results was initiated by the user by  
-     *             entering a different search criteria or by viewing the results list in a different
-     *             size mode, but we keep the page selection the same if the results got updated on
-     *             their own while the user was browsing through the result pages.
+     * @flags: Flags controlling redisplay behavior as follows:
+     *  SUBSEARCH - Indicates that the current _search is a superstring of the previous
+     *  one, which implies we only need to re-search through previous results.
+     *  FULL - Indicates that we need recreate all displayed items.
+     *  IMMEDIATE - Do the full redisplay even if we're not mapped.  This is useful
+     *  if you want to get the number of matched items and show/hide a section based on
+     *  that number.
      */
-    _redisplay: function(resetPage) {
-        this._refreshCache();
-        if (!this._filterActive())
-            this._setDefaultList();
-        else
-            this._doSearchFilter();
+    _redisplay: function(flags) {
+        let immediate = (flags & RedisplayFlags.IMMEDIATE) != 0;
+        if (!immediate && !this.actor.mapped) {
+            this._pendingRedisplay |= flags;
+            return;
+        }
 
-        if (resetPage)
-            this._list.page = 0;
+        let isSubSearch = (flags & RedisplayFlags.SUBSEARCH) != 0;
+        let fullReload = (flags & RedisplayFlags.FULL) != 0;
 
-        this._displayMatchedItems(true);
+        let hadSelected = this.hasSelected();
+        this.unsetSelected();
+
+        if (!this._initialLoadComplete)
+            fullReload = true;
+
+        if (!this._refreshCache())
+            fullReload = true;
+
+        if (fullReload) {
+            this._recreateDisplayItems();
+            this._initialLoadComplete = true;
+        }
+
+        if (isSubSearch) {
+            this._redisplaySubSearch();
+        } else {
+            this._redisplayReordering();
+        }
+
+        if (hadSelected) {
+            this._selectedIndex = -1;
+            this.selectFirstItem();
+        }
 
         this.emit('redisplayed');
     },
 
     //// Pure virtual protected methods ////
- 
+
     // Performs the steps needed to have the latest information about the items.
+    // Implementation should return %true if we are up to date, and %false
+    // if a full reload occurred.
     _refreshCache: function() {
         throw new Error("Not implemented");
     },
@@ -624,10 +608,10 @@ GenericDisplay.prototype = {
         throw new Error("Not implemented");
     },
 
-	// Compares items associated with the item ids based on the order in which the
-	// items should be displayed.
-	// Intended to be used as a compareFunction for array.sort().
-	// Returns an integer value indicating the result of the comparison.
+    // Compares items associated with the item ids based on the order in which the
+    // items should be displayed.
+    // Intended to be used as a compareFunction for array.sort().
+    // Returns an integer value indicating the result of the comparison.
     _compareItems: function(itemIdA, itemIdB) {
         throw new Error("Not implemented");
     },
@@ -645,108 +629,50 @@ GenericDisplay.prototype = {
 
     //// Private methods ////
 
-    _getSearchMatchedItems: function() {
-        let matchedItemsForSearch = {};
+    _getItemSearchScore: function(itemId, terms) {
+        let item = this._allItems[itemId];
+        let score = 0;
+        for (let i = 0; i < terms.length; i++) {
+            let term = terms[i];
+            if (this._isInfoMatching(item, term)) {
+                score++;
+            }
+        }
+        return score;
+    },
+
+    _getSearchMatchedItems: function(isSubSearch) {
         // Break the search up into terms, and search for each
         // individual term.  Keep track of the number of terms
         // each item matched.
         let terms = this._search.split(/\s+/);
-        for (let i = 0; i < terms.length; i++) {
-            let term = terms[i];
-            for (itemId in this._allItems) {
-                let item = this._allItems[itemId];
-                if (this._isInfoMatching(item, term)) {
-                    let count = matchedItemsForSearch[itemId];
-                    if (!count)
-                        count = 0;
-                    count += 1;
-                    matchedItemsForSearch[itemId] = count;
-                }
-            }
-        }
-        return matchedItemsForSearch;
-    },
+        let matchScores = {};
 
-    // Applies the search string to the list of items to find matches,
-    // and displays the matching items.
-    _doSearchFilter: function() {
-        let matchedItemsForSearch;
-
-        if (this._filterActive()) {
-            matchedItemsForSearch = this._getSearchMatchedItems();
-        } else {
-            matchedItemsForSearch = {};
-            for (let itemId in this._allItems) {
-                matchedItemsForSearch[itemId] = 1;
-            }
-        }
-
-        this._matchedItems = [];
-        for (itemId in matchedItemsForSearch) {
-            this._matchedItems.push(itemId);
-        }
-        this._matchedItems.sort(Lang.bind(this, function (a, b) {
-            let countA = matchedItemsForSearch[a];
-            let countB = matchedItemsForSearch[b];
-            if (countA > countB)
-                return -1;
-            else if (countA < countB)
-                return 1;
-            else
-                return this._compareItems(a, b);
-        }));
-    },
-
-    /*
-     * Updates the display control to reflect the matched items set and the page selected.
-     *
-     * resetDisplayControl - indicates if the display control should be re-created because 
-     *                       the results or the space allocated for them changed. If it's false,
-     *                       the existing display control is used and only the page links are
-     *                       updated to reflect the current page selection.
-     */
-    _updateDisplayControl: function(resetDisplayControl) {
-        if (resetDisplayControl) {
-            this.displayControl.remove_all();
-            let nPages = this._list.n_pages;
-            let pageNumber = this._list.page;
-            for (let i = 0; i < nPages; i++) {
-                let pageControl = new Link.Link({ color: (i == pageNumber) ? DISPLAY_CONTROL_SELECTED_COLOR : ITEM_DISPLAY_DESCRIPTION_COLOR,
-                                                  font_name: "Sans Bold 16px",
-                                                  text: (i+1) + "",
-                                                  reactive: (i == pageNumber) ? false : true});
-                this.displayControl.append(pageControl.actor, Big.BoxPackFlags.NONE);
-
-                // we use pageNumberLocalScope to get the page number right in the callback function
-                let pageNumberLocalScope = i;
-                pageControl.connect('clicked',
-                                    Lang.bind(this,
-                                              function(o, event) {
-                                                  this.displayPage(pageNumberLocalScope);
-                                              }));
+        if (isSubSearch) {
+            for (let i = 0; i < this._matchedItemKeys.length; i++) {
+                let itemId = this._matchedItemKeys[i];
+                let score = this._getItemSearchScore(itemId, terms);
+                if (score > 0)
+                    matchScores[itemId] = score;
             }
         } else {
-            let pageControlActors = this.displayControl.get_children();
-            for (let i = 0; i < pageControlActors.length; i++) {
-                let pageControlActor = pageControlActors[i];
-                if (i == this._list.page) {
-                    pageControlActor.color =  DISPLAY_CONTROL_SELECTED_COLOR;
-                    pageControlActor.reactive = false;
-                } else {
-                    pageControlActor.color =  ITEM_DISPLAY_DESCRIPTION_COLOR;
-                    pageControlActor.reactive = true;
-                }
-            } 
-        }
-        if (this.hasSelected()) {
-            this.selectFirstItem();
+            for (let itemId in this._displayedItems) {
+                let score = this._getItemSearchScore(itemId, terms);
+                if (score > 0)
+                    matchScores[itemId] = score;
+            }
         }
+        return matchScores;
     },
 
     // Returns a display item based on its index in the ordering of the
     // display children.
     _findDisplayedByIndex: function(index) {
-        let actor = this._list.get_displayed_actor(index);
+        let actor;
+        if (this.actor instanceof Shell.OverflowList)
+            actor = this.actor.get_displayed_actor(index);
+        else
+            actor = this._list.get_children()[index];
         return this._findDisplayedByActor(actor);
     },
 
@@ -778,6 +704,20 @@ GenericDisplay.prototype = {
         let item = this._findDisplayedByIndex(index);
         item.markSelected(true);
         this.emit('selected');
+    },
+
+    _getVisibleCount: function() {
+        if (this.actor instanceof Shell.OverflowList)
+            return this._list.displayed_count;
+        return this._list.get_n_children();
+    },
+
+    _onMappedNotify: function () {
+        let mapped = this.actor.mapped;
+        if (mapped && this._pendingRedisplay > RedisplayFlags.NONE)
+            this._redisplay(this._pendingRedisplay);
+
+        this._pendingRedisplay = RedisplayFlags.NONE;
     }
 };
 
