@@ -4,26 +4,20 @@ const Big = imports.gi.Big;
 const Clutter = imports.gi.Clutter;
 const Gio = imports.gi.Gio;
 const Pango = imports.gi.Pango;
+const St = imports.gi.St;
 const Shell = imports.gi.Shell;
 const Signals = imports.signals;
 const Lang = imports.lang;
 const Mainloop = imports.mainloop;
+const Gettext = imports.gettext.domain('gnome-shell');
+const _ = Gettext.gettext;
 
+const ExtensionSystem = imports.ui.extensionSystem;
+const Link = imports.ui.link;
 const Tweener = imports.ui.tweener;
 const Main = imports.ui.main;
 
-const LG_BORDER_COLOR = new Clutter.Color();
-LG_BORDER_COLOR.from_pixel(0x0000aca0);
-const LG_BACKGROUND_COLOR = new Clutter.Color();
-LG_BACKGROUND_COLOR.from_pixel(0x000000d5);
-const GREY = new Clutter.Color();
-GREY.from_pixel(0xAFAFAFFF);
-const MATRIX_GREEN = new Clutter.Color();
-MATRIX_GREEN.from_pixel(0x88ff66ff);
-// FIXME pull from GConf
-const MATRIX_FONT = 'Monospace 10';
-
-                    /* Imports...feel free to add here as needed */
+/* Imports...feel free to add here as needed */
 var commandHeader = "const Clutter = imports.gi.Clutter; " +
                     "const GLib = imports.gi.GLib; " +
                     "const Gtk = imports.gi.Gtk; " +
@@ -47,32 +41,41 @@ function Notebook() {
 
 Notebook.prototype = {
     _init: function() {
-        this.actor = new Big.Box();
+        this.actor = new St.BoxLayout({ vertical: true });
 
-        this.tabControls = new Big.Box({ orientation: Big.BoxOrientation.HORIZONTAL,
-                                         spacing: 4, padding: 2 });
+        this.tabControls = new St.BoxLayout({ style_class: "labels" });
 
         this._selectedIndex = -1;
         this._tabs = [];
     },
 
     appendPage: function(name, child) {
-        let labelOuterBox = new Big.Box({ padding: 2 });
-        let labelBox = new Big.Box({ padding: 2, border_color: MATRIX_GREEN,
-                                     reactive: true });
-        labelOuterBox.append(labelBox, Big.BoxPackFlags.NONE);
-        let label = new Clutter.Text({ color: MATRIX_GREEN,
-                                       font_name: MATRIX_FONT,
-                                       text: name });
-        labelBox.connect('button-press-event', Lang.bind(this, function () {
+        let labelBox = new St.BoxLayout({ style_class: "notebook-tab" });
+        let label = new St.Button({ label: name });
+        label.connect('clicked', Lang.bind(this, function () {
             this.selectChild(child);
             return true;
         }));
-        labelBox.append(label, Big.BoxPackFlags.EXPAND);
-        this._tabs.push([child, labelBox]);
-        child.hide();
-        this.actor.append(child, Big.BoxPackFlags.EXPAND);
-        this.tabControls.append(labelOuterBox, Big.BoxPackFlags.NONE);
+        labelBox.add(label, { expand: true });
+        this.tabControls.add(labelBox);
+
+        let scrollview = new St.ScrollView({ x_fill: true, y_fill: true });
+        scrollview.get_hscroll_bar().hide();
+        scrollview.add_actor(child);
+
+        let tabData = { child: child,
+                        labelBox: labelBox,
+                        label: label,
+                        scrollView: scrollview,
+                        _scrollToBottom: false };
+        this._tabs.push(tabData);
+        scrollview.hide();
+        this.actor.add(scrollview, { expand: true });
+
+        let vAdjust = scrollview.vscroll.adjustment;
+        vAdjust.connect('changed', Lang.bind(this, function () { this._onAdjustScopeChanged(tabData); }));
+        vAdjust.connect('notify::value', Lang.bind(this, function() { this._onAdjustValueChanged(tabData); }));
+
         if (this._selectedIndex == -1)
             this.selectIndex(0);
     },
@@ -80,10 +83,9 @@ Notebook.prototype = {
     _unselect: function() {
         if (this._selectedIndex < 0)
             return;
-        let [child, labelBox] = this._tabs[this._selectedIndex];
-        labelBox.padding = 2;
-        labelBox.border = 0;
-        child.hide();
+        let tabData = this._tabs[this._selectedIndex];
+        tabData.labelBox.set_style_pseudo_class(null);
+        tabData.scrollView.hide();
         this._selectedIndex = -1;
     },
 
@@ -95,12 +97,11 @@ Notebook.prototype = {
             this.emit('selection', null);
             return;
         }
-        let [child, labelBox] = this._tabs[index];
-        labelBox.padding = 1;
-        labelBox.border = 1;
-        child.show();
+        let tabData = this._tabs[index];
+        tabData.labelBox.set_style_pseudo_class('selected');
+        tabData.scrollView.show();
         this._selectedIndex = index;
-        this.emit('selection', child);
+        this.emit('selection', tabData.child);
     },
 
     selectChild: function(child) {
@@ -108,13 +109,32 @@ Notebook.prototype = {
             this.selectIndex(-1);
         else {
             for (let i = 0; i < this._tabs.length; i++) {
-                let [tabChild, labelBox] = this._tabs[i];
-                if (tabChild == child) {
+                let tabData = this._tabs[i];
+                if (tabData.child == child) {
                     this.selectIndex(i);
                     return;
                 }
             }
         }
+    },
+
+    scrollToBottom: function(index) {
+        let tabData = this._tabs[index];
+        tabData._scrollToBottom = true;
+
+    },
+
+    _onAdjustValueChanged: function (tabData) {
+        let vAdjust = tabData.scrollView.vscroll.adjustment;
+        if (vAdjust.value < (vAdjust.upper - vAdjust.lower - 0.5))
+            tabData._scrolltoBottom = false;
+    },
+
+    _onAdjustScopeChanged: function (tabData) {
+        if (!tabData._scrollToBottom)
+            return;
+        let vAdjust = tabData.scrollView.vscroll.adjustment;
+        vAdjust.value = vAdjust.upper - vAdjust.page_size;
     }
 }
 Signals.addSignalMethods(Notebook.prototype);
@@ -130,20 +150,19 @@ Result.prototype = {
 
         this.actor = new Big.Box();
 
-        let cmdTxt = new Clutter.Text({ color: MATRIX_GREEN,
-                                        font_name: MATRIX_FONT,
-                                        ellipsize: Pango.EllipsizeMode.END,
-                                        text: command });
+        let cmdTxt = new St.Label({ text: command });
+        cmdTxt.ellipsize = Pango.EllipsizeMode.END;
+
         this.actor.append(cmdTxt, Big.BoxPackFlags.NONE);
-        let resultTxt = new Clutter.Text({ color: MATRIX_GREEN,
-                                           font_name: MATRIX_FONT,
-                                           ellipsize: Pango.EllipsizeMode.END,
-                                           text: "r(" + index + ") = " + o });
+        let resultTxt = new St.Label({ text: "r(" + index + ") = " + o });
+        resultTxt.ellipsize = Pango.EllipsizeMode.END;
+
         this.actor.append(resultTxt, Big.BoxPackFlags.NONE);
-        let line = new Big.Box({ border_color: GREY,
-                                 border_bottom: 1,
-                                 height: 8 });
-        this.actor.append(line, Big.BoxPackFlags.NONE);
+        let line = new Clutter.Rectangle({ name: "Separator",
+                                           height: 1 });
+        let padBin = new St.Bin({ name: "Separator", x_fill: true, y_fill: true });
+        padBin.add_actor(line);
+        this.actor.append(padBin, Big.BoxPackFlags.NONE);
     }
 }
 
@@ -158,17 +177,14 @@ ActorHierarchy.prototype = {
 
         this._parentList = [];
 
-        this.actor = new Big.Box({ spacing: 4,
-                                   border: 1,
-                                   padding: 4,
-                                   border_color: GREY });
+        this.actor = new St.BoxLayout({ name: "ActorHierarchy", vertical: true });
     },
 
     setTarget: function(actor) {
         this._previousTarget = this._target;
         this.target = actor;
 
-        this.actor.remove_all();
+        this.actor.get_children().forEach(function (child) { child.destroy(); });
 
         if (!(actor instanceof Clutter.Actor))
             return;
@@ -181,11 +197,9 @@ ActorHierarchy.prototype = {
         while ((parent = parent.get_parent()) != null) {
             this._parentList.push(parent);
 
-            let link = new Clutter.Text({ color: MATRIX_GREEN,
-                                          font_name: MATRIX_FONT,
-                                          reactive: true,
-                                          text: "" + parent });
-            this.actor.append(link, Big.BoxPackFlags.IF_FITS);
+            let link = new St.Label({ reactive: true,
+                                      text: "" + parent });
+            this.actor.add_actor(link);
             let parentTarget = parent;
             link.connect('button-press-event', Lang.bind(this, function () {
                 this._selectByActor(parentTarget);
@@ -214,16 +228,13 @@ PropertyInspector.prototype = {
 
         this._parentList = [];
 
-        this.actor = new Big.Box({ spacing: 4,
-                                   border: 1,
-                                   padding: 4,
-                                   border_color: GREY });
+        this.actor = new St.BoxLayout({ name: "PropertyInspector", vertical: true });
     },
 
     setTarget: function(actor) {
         this.target = actor;
 
-        this.actor.remove_all();
+        this.actor.get_children().forEach(function (child) { child.destroy(); });
 
         for (let propName in actor) {
             let valueStr;
@@ -233,11 +244,9 @@ PropertyInspector.prototype = {
                 valueStr = '<error>';
             }
             let propText = propName + ": " + valueStr;
-            let propDisplay = new Clutter.Text({ color: MATRIX_GREEN,
-                                                 font_name: MATRIX_FONT,
-                                                 reactive: true,
-                                                 text: propText });
-            this.actor.append(propDisplay, Big.BoxPackFlags.IF_FITS);
+            let propDisplay = new St.Label({ reactive: true,
+                                             text: propText });
+            this.actor.add_actor(propDisplay);
         }
     }
 }
@@ -249,20 +258,17 @@ function Inspector() {
 Inspector.prototype = {
     _init: function() {
         let width = 150;
-        let eventHandler = new Big.Box({ background_color: LG_BACKGROUND_COLOR,
-                                         border: 1,
-                                         border_color: LG_BORDER_COLOR,
-                                         corner_radius: 4,
-                                         y: global.stage.height/2,
-                                         reactive: true
-                                      });
+        let primary = global.get_primary_monitor();
+        let eventHandler = new St.BoxLayout({ name: "LookingGlassDialog",
+                                              vertical: false,
+                                              y: primary.y + Math.floor(primary.height / 2),
+                                              reactive: true });
         eventHandler.connect('notify::allocation', Lang.bind(this, function () {
-            eventHandler.x = Math.floor((global.stage.width)/2 - (eventHandler.width)/2);
+            eventHandler.x = primary.x + Math.floor((primary.width - eventHandler.width) / 2);
         }));
         global.stage.add_actor(eventHandler);
-        let displayText = new Clutter.Text({ color: MATRIX_GREEN,
-                                             font_name: MATRIX_FONT, text: '' });
-        eventHandler.append(displayText, Big.BoxPackFlags.EXPAND);
+        let displayText = new St.Label();
+        eventHandler.add(displayText, { expand: true });
 
         let borderPaintTarget = null;
         let borderPaintId = null;
@@ -302,6 +308,135 @@ Inspector.prototype = {
 
 Signals.addSignalMethods(Inspector.prototype);
 
+function ErrorLog() {
+    this._init();
+}
+
+ErrorLog.prototype = {
+    _init: function() {
+        this.actor = new St.BoxLayout();
+        this.text = new St.Label();
+        this.actor.add(this.text);
+        this.text.clutter_text.line_wrap = true;
+        this.actor.connect('notify::mapped', Lang.bind(this, this._renderText));
+    },
+
+    _formatTime: function(d){
+        function pad(n) { return n < 10 ? '0' + n : n };
+        return d.getUTCFullYear()+'-'
+            + pad(d.getUTCMonth()+1)+'-'
+            + pad(d.getUTCDate())+'T'
+            + pad(d.getUTCHours())+':'
+            + pad(d.getUTCMinutes())+':'
+            + pad(d.getUTCSeconds())+'Z'
+    },
+
+    _renderText: function() {
+        if (!this.actor.mapped)
+            return;
+        let text = this.text.text;
+        let stack = Main._getAndClearErrorStack();
+        for (let i = 0; i < stack.length; i++) {
+            let logItem = stack[i];
+            text += logItem.category + " t=" + this._formatTime(new Date(logItem.timestamp)) + " " + logItem.message + "\n";
+        }
+        this.text.text = text;
+    }
+}
+
+function Extensions() {
+    this._init();
+}
+
+Extensions.prototype = {
+    _init: function() {
+        this.actor = new St.BoxLayout({ vertical: true,
+                                        name: 'lookingGlassExtensions' });
+        this._noExtensions = new St.Label({ style_class: 'lg-extensions-none',
+                                             text: _("No extensions installed") });
+        this._extensionsList = new St.BoxLayout({ vertical: true,
+                                                  style_class: 'lg-extensions-list' });
+        this.actor.add(this._extensionsList);
+        this._loadExtensionList();
+    },
+
+    _loadExtensionList: function() {
+        let extensions = ExtensionSystem.extensionMeta;
+        let totalExtensions = 0;
+        for (let uuid in extensions) {
+            let extensionDisplay = this._createExtensionDisplay(extensions[uuid]);
+            this._extensionsList.add(extensionDisplay);
+            totalExtensions++;
+        }
+        if (totalExtensions == 0) {
+            this._extensionsList.add(this._noExtensions);
+        }
+    },
+
+    _onViewSource: function (actor) {
+        let meta = actor._extensionMeta;
+        let file = Gio.file_new_for_path(meta.path);
+        let uri = file.get_uri();
+        Gio.app_info_launch_default_for_uri(uri, global.create_app_launch_context());
+        Main.lookingGlass.close();
+    },
+
+    _onWebPage: function (actor) {
+        let meta = actor._extensionMeta;
+        Gio.app_info_launch_default_for_uri(meta.url, global.create_app_launch_context());
+        Main.lookingGlass.close();
+    },
+
+    _stateToString: function(extensionState) {
+        switch (extensionState) {
+            case ExtensionSystem.ExtensionState.ENABLED:
+                return _("Enabled");
+            case ExtensionSystem.ExtensionState.DISABLED:
+                return _("Disabled");
+            case ExtensionSystem.ExtensionState.ERROR:
+                return _("Error");
+            case ExtensionSystem.ExtensionState.OUT_OF_DATE:
+                return _("Out of date");
+        }
+        return "Unknown"; // Not translated, shouldn't appear
+    },
+
+    _createExtensionDisplay: function(meta) {
+        let box = new St.BoxLayout({ style_class: 'lg-extension', vertical: true });
+        let name = new St.Label({ style_class: 'lg-extension-name',
+                                   text: meta.name });
+        box.add(name, { expand: true });
+        let description = new St.Label({ style_class: 'lg-extension-description',
+                                         text: meta.description });
+        box.add(description, { expand: true });
+
+        let metaBox = new St.BoxLayout();
+        box.add(metaBox);
+        let stateString = this._stateToString(meta.state);
+        let state = new St.Label({ style_class: 'lg-extension-state',
+                                   text: this._stateToString(meta.state) });
+
+        let actionsContainer = new St.Bin({ x_align: St.Align.END });
+        metaBox.add(actionsContainer);
+        let actionsBox = new St.BoxLayout({ style_class: 'lg-extension-actions' });
+        actionsContainer.set_child(actionsBox);
+
+        let viewsource = new Link.Link({ label: _("View Source") });
+        viewsource.actor._extensionMeta = meta;
+        viewsource.actor.connect('clicked', Lang.bind(this, this._onViewSource));
+        actionsBox.add(viewsource.actor);
+
+        if (meta.url) {
+            let webpage = new Link.Link({ label: _("Web Page") });
+            webpage.actor._extensionMeta = meta;
+            webpage.actor.connect('clicked', Lang.bind(this, this._onWebPage));
+            actionsBox.add(webpage.actor);
+        }
+
+        return box;
+    }
+};
+
 function LookingGlass() {
     this._init();
 }
@@ -314,6 +449,10 @@ LookingGlass.prototype = {
         this._savedText = null;
         this._historyNavIndex = -1;
         this._history = [];
+        this._borderPaintTarget = null;
+        this._borderPaintId = 0;
+        this._borderDestroyId = 0;
+
         this._readHistory();
 
         this._open = false;
@@ -321,29 +460,26 @@ LookingGlass.prototype = {
         this._offset = 0;
         this._results = [];
 
-        // TODO replace with scrolling or something better
-        this._maxItems = 10;
+        // Sort of magic, but...eh.
+        this._maxItems = 150;
+
+        this.actor = new St.BoxLayout({ name: "LookingGlassDialog",
+                                        vertical: true,
+                                        visible: false });
+
+        let gconf = Shell.GConf.get_default();
+        gconf.watch_directory("/desktop/gnome/interface");
+        gconf.connect("changed::/desktop/gnome/interface/monospace_font_name",
+                      Lang.bind(this, this._updateFont));
+        this._updateFont();
 
-        this.actor = new Big.Box({ background_color: LG_BACKGROUND_COLOR,
-                                   border: 1,
-                                   border_color: LG_BORDER_COLOR,
-                                   corner_radius: 4,
-                                   padding_top: 8,
-                                   padding_left: 4,
-                                   padding_right: 4,
-                                   padding_bottom: 4,
-                                   spacing: 4,
-                                   visible: false
-                                });
         global.stage.add_actor(this.actor);
 
-        let toolbar = new Big.Box({ orientation: Big.BoxOrientation.HORIZONTAL,
-                                    border: 1, border_color: GREY,
-                                    corner_radius: 4 });
-        this.actor.append(toolbar, Big.BoxPackFlags.NONE);
+        let toolbar = new St.BoxLayout({ name: "Toolbar" });
+        this.actor.add_actor(toolbar);
         let inspectIcon = Shell.TextureCache.get_default().load_gicon(new Gio.ThemedIcon({ name: 'gtk-color-picker' }),
                                                                       24);
-        toolbar.append(inspectIcon, Big.BoxPackFlags.NONE);
+        toolbar.add_actor(inspectIcon);
         inspectIcon.reactive = true;
         inspectIcon.connect('button-press-event', Lang.bind(this, function () {
             let inspector = new Inspector();
@@ -361,31 +497,27 @@ LookingGlass.prototype = {
         }));
 
         let notebook = new Notebook();
-        this.actor.append(notebook.actor, Big.BoxPackFlags.EXPAND);
-        toolbar.append(notebook.tabControls, Big.BoxPackFlags.END);
+        this._notebook = notebook;
+        this.actor.add(notebook.actor, { expand: true });
 
-        this._evalBox = new Big.Box({ orientation: Big.BoxOrientation.VERTICAL,
-                                      spacing: 4 });
+        let emptyBox = new St.Bin();
+        toolbar.add(emptyBox, { expand: true });
+        toolbar.add_actor(notebook.tabControls);
+
+        this._evalBox = new St.BoxLayout({ name: "EvalBox", vertical: true });
         notebook.appendPage('Evaluator', this._evalBox);
 
         this._resultsArea = new Big.Box({ orientation: Big.BoxOrientation.VERTICAL,
                                           spacing: 4 });
-        this._evalBox.append(this._resultsArea, Big.BoxPackFlags.EXPAND);
+        this._evalBox.add(this._resultsArea, { expand: true });
 
         let entryArea = new Big.Box({ orientation: Big.BoxOrientation.HORIZONTAL });
-        this._evalBox.append(entryArea, Big.BoxPackFlags.NONE);
+        this._evalBox.add_actor(entryArea);
 
-        let label = new Clutter.Text({ color: MATRIX_GREEN,
-                                       font_name: MATRIX_FONT,
-                                       text: 'js>>> ' });
+        let label = new St.Label({ text: 'js>>> ' });
         entryArea.append(label, Big.BoxPackFlags.NONE);
 
-        this._entry = new Clutter.Text({ color: MATRIX_GREEN,
-                                         font_name: MATRIX_FONT,
-                                         editable: true,
-                                         activatable: true,
-                                         singleLineMode: true,
-                                         text: ''});
+        this._entry = new St.Entry();
         /* unmapping the edit box will un-focus it, undo that */
         notebook.connect('selection', Lang.bind(this, function (nb, child) {
             if (child == this._evalBox)
@@ -403,7 +535,13 @@ LookingGlass.prototype = {
             notebook.selectIndex(0);
         }));
 
-        this._entry.connect('activate', Lang.bind(this, function (o, e) {
+        this._errorLog = new ErrorLog();
+        notebook.appendPage('Errors', this._errorLog.actor);
+
+        this._extensions = new Extensions();
+        notebook.appendPage('Extensions', this._extensions.actor);
+
+        this._entry.clutter_text.connect('activate', Lang.bind(this, function (o, e) {
             let text = o.get_text();
             // Ensure we don't get newlines in the command; the history file is
             // newline-separated.
@@ -416,12 +554,9 @@ LookingGlass.prototype = {
             this._historyNavIndex = -1;
             return true;
         }));
-        this._entry.connect('key-press-event', Lang.bind(this, function(o, e) {
+        this._entry.clutter_text.connect('key-press-event', Lang.bind(this, function(o, e) {
             let symbol = e.get_key_symbol();
-            if (symbol == Clutter.Escape) {
-                this.close();
-                return true;
-            } else if (symbol == Clutter.Up) {
+            if (symbol == Clutter.Up) {
                 if (this._historyNavIndex >= this._history.length - 1)
                     return true;
                 this._historyNavIndex++;
@@ -446,6 +581,19 @@ LookingGlass.prototype = {
         }));
     },
 
+    _updateFont: function() {
+        let gconf = Shell.GConf.get_default();
+        let fontName = gconf.get_string("/desktop/gnome/interface/monospace_font_name");
+        // This is mishandled by the scanner - should by Pango.FontDescription_from_string(fontName);
+        // https://bugzilla.gnome.org/show_bug.cgi?id=595889
+        let fontDesc = Pango.Font.description_from_string(fontName);
+        // We ignore everything but size and style; you'd be crazy to set your system-wide
+        // monospace font to be bold/oblique/etc. Could easily be added here.
+        this.actor.style =
+            'font-size: ' + fontDesc.get_size() / 1024. + (fontDesc.get_size_is_absolute() ? 'px' : 'pt') + ';'
+            + 'font-family: "' + fontDesc.get_family() + '";';
+    },
+
     _readHistory: function () {
         if (!this._historyFile.query_exists(null))
             return;
@@ -475,6 +623,18 @@ LookingGlass.prototype = {
         this._results.push(result);
         this._resultsArea.append(result.actor, Big.BoxPackFlags.NONE);
         this._propInspector.setTarget(obj);
+        if (this._borderPaintTarget != null) {
+            this._borderPaintTarget.disconnect(this._borderPaintId);
+            this._borderPaintTarget = null;
+        }
+        if (obj instanceof Clutter.Actor) {
+            this._borderPaintTarget = obj;
+            this._borderPaintId = Shell.add_hook_paint_red_border(obj);
+            this._borderDestroyId = obj.connect('destroy', Lang.bind(this, function () {
+                this._borderDestroyId = 0;
+                this._borderPaintTarget = null;
+            }));
+        }
         let children = this._resultsArea.get_children();
         if (children.length > this._maxItems) {
             this._results.shift();
@@ -482,6 +642,9 @@ LookingGlass.prototype = {
             this._offset++;
         }
         this._it = obj;
+
+        // Scroll to bottom
+        this._notebook.scrollToBottom(0);
     },
 
     _evaluate : function(command) {
@@ -518,11 +681,11 @@ LookingGlass.prototype = {
     },
 
     _resizeTo: function(actor) {
-        let stage = global.stage;
-        let myWidth = stage.width * 0.7;
-        let myHeight = stage.height * 0.7;
+        let primary = global.get_primary_monitor();
+        let myWidth = primary.width * 0.7;
+        let myHeight = primary.height * 0.7;
         let [srcX, srcY] = actor.get_transformed_position();
-        this.actor.x = srcX + (stage.width-myWidth)/2;
+        this.actor.x = srcX + (primary.width - myWidth) / 2;
         this._hiddenY = srcY + actor.height - myHeight - 4; // -4 to hide the top corners
         this._targetY = this._hiddenY + myHeight;
         this.actor.y = this._hiddenY;
@@ -538,17 +701,32 @@ LookingGlass.prototype = {
         this._resizeTo(actor);
     },
 
+    // Handle key events which are relevant for all tabs of the LookingGlass
+    _globalKeyPressEvent : function(actor, event) {
+        let symbol = event.get_key_symbol();
+        if (symbol == Clutter.Escape) {
+            this.close();
+            return true;
+        }
+        return false;
+    },
+
     open : function() {
         if (this._open)
             return;
 
+        if (!Main.pushModal(this.actor))
+            return;
+
+        this._keyPressEventId = global.stage.connect('key-press-event',
+            Lang.bind(this, this._globalKeyPressEvent));
+
         this.actor.show();
         this.actor.lower(Main.chrome.actor);
         this._open = true;
 
         Tweener.removeTweens(this.actor);
 
-        Main.pushModal(this.actor);
         global.stage.set_key_focus(this._entry);
 
         Tweener.addTween(this.actor, { time: 0.5,
@@ -561,10 +739,19 @@ LookingGlass.prototype = {
         if (!this._open)
             return;
 
+        if (this._keyPressEventId)
+            global.stage.disconnect(this._keyPressEventId);
+
         this._historyNavIndex = -1;
         this._open = false;
         Tweener.removeTweens(this.actor);
 
+        if (this._borderPaintTarget != null) {
+            this._borderPaintTarget.disconnect(this._borderPaintId);
+            this._borderPaintTarget.disconnect(this._borderDestroyId);
+            this._borderPaintTarget = null;
+        }
+
         Main.popModal(this.actor);
 
         Tweener.addTween(this.actor, { time: 0.5,
