@@ -27,13 +27,13 @@ WidgetBox.prototype = {
     _init: function(widget, expanded) {
         this.state = expanded ? Widget.STATE_EXPANDED : Widget.STATE_COLLAPSED;
 
-	if (widget instanceof Widget.Widget) {
-	    this._widget = widget;
+        if (widget instanceof Widget.Widget) {
+            this._widget = widget;
             this._widget.state = this.state;
-	} else {
-	    let ctor = this._ctorFromName(widget);
+        } else {
+            let ctor = this._ctorFromName(widget);
             this._widget = new ctor(this.state);
-	}
+        }
 
         if (!this._widget.actor)
             throw new Error("widget has no actor");
@@ -331,7 +331,7 @@ WidgetBox.prototype = {
                                        onCompleteScope: this });
         this.state = this._widget.state = Widget.STATE_POPPING_OUT;
 
-        Main.chrome.addInputRegionActor(this._hbox);
+        Main.chrome.trackActor(this._hbox, { affectsStruts: false });
     },
 
     _popOutComplete: function() {
@@ -370,7 +370,21 @@ WidgetBox.prototype = {
         this._egroup.hide();
         this._ebox.x = 0;
 
-        Main.chrome.removeInputRegionActor(this._hbox);
+        Main.chrome.untrackActor(this._hbox);
+    },
+
+    hide: function() {
+        if (this.state == Widget.STATE_COLLAPSED)
+            Tweener.addTween(this._cbox, { x: -Widget.COLLAPSED_WIDTH,
+                                           time: ANIMATION_TIME / 2,
+                                           transition: "easeOutQuad" });
+    },
+
+    restore: function() {
+        if (this.state == Widget.STATE_COLLAPSED)
+            Tweener.addTween(this._cbox, { x: 0,
+                                           time: ANIMATION_TIME / 2,
+                                           transition: "easeOutQuad" });
     },
 
     destroy: function() {
