@@ -21,14 +21,27 @@ struct _StThemeNode {
   PangoFontDescription *font_desc;
 
   ClutterColor background_color;
+  /* If gradient is set, then background_color is the gradient start */
+  StGradientType background_gradient_type;
+  ClutterColor background_gradient_end;
+
   ClutterColor foreground_color;
   ClutterColor border_color[4];
+
   double border_width[4];
   double border_radius[4];
   guint padding[4];
 
+  int width;
+  int height;
+  int min_width;
+  int min_height;
+  int max_width;
+  int max_height;
+
   char *background_image;
   StBorderImage *border_image;
+  StShadow *shadow;
 
   GType element_type;
   char *element_id;
@@ -43,10 +56,11 @@ struct _StThemeNode {
   CRDeclaration *inline_properties;
 
   guint properties_computed : 1;
-  guint borders_computed : 1;
+  guint geometry_computed : 1;
   guint background_computed : 1;
   guint foreground_computed : 1;
   guint border_image_computed : 1;
+  guint shadow_computed : 1;
   guint link_type : 2;
 };
 
@@ -108,6 +122,12 @@ st_theme_node_finalize (GObject *object)
       node->border_image = NULL;
     }
 
+  if (node->shadow)
+    {
+      st_shadow_free (node->shadow);
+      node->shadow = NULL;
+    }
+
   if (node->background_image)
     g_free (node->background_image);
 
@@ -452,7 +472,7 @@ get_color_from_term (StThemeNode  *node,
  *   parent's parent, and so forth. Note that if the property has a
  *   value of 'inherit' it will be inherited even if %FALSE is passed
  *   in for @inherit; this only affects the default behavior for inheritance.
- * @color: (out): location to store the color that was determined.
+ * @color: location to store the color that was determined.
  *   If the property is not found, the value in this location
  *   will not be changed.
  *
@@ -992,18 +1012,22 @@ do_padding_property_term (StThemeNode *node,
                           gboolean     bottom)
 {
   gdouble value;
+  int int_value;
 
   if (get_length_from_term (node, term, FALSE, &value) != VALUE_FOUND)
     return;
 
+  /* Round the value */
+  int_value = (int) (0.5 + value);
+
   if (left)
-    node->padding[ST_SIDE_LEFT] = value;
+    node->padding[ST_SIDE_LEFT] = int_value;
   if (right)
-    node->padding[ST_SIDE_RIGHT] = value;
+    node->padding[ST_SIDE_RIGHT] = int_value;
   if (top)
-    node->padding[ST_SIDE_TOP] = value;
+    node->padding[ST_SIDE_TOP] = int_value;
   if (bottom)
-    node->padding[ST_SIDE_BOTTOM] = value;
+    node->padding[ST_SIDE_BOTTOM] = int_value;
 }
 
 static void
@@ -1066,14 +1090,25 @@ do_padding_property (StThemeNode   *node,
 }
 
 static void
-ensure_borders (StThemeNode *node)
+do_size_property (StThemeNode   *node,
+                  CRDeclaration *decl,
+		  int           *node_value)
+{
+  gdouble value;
+
+  if (get_length_from_term (node, decl->value, FALSE, &value) == VALUE_FOUND)
+    *node_value = (int) (0.5 + value);
+}
+
+static void
+ensure_geometry (StThemeNode *node)
 {
   int i, j;
 
-  if (node->borders_computed)
+  if (node->geometry_computed)
     return;
 
-  node->borders_computed = TRUE;
+  node->geometry_computed = TRUE;
 
   ensure_properties (node);
 
@@ -1083,6 +1118,13 @@ ensure_borders (StThemeNode *node)
       node->border_color[j] = TRANSPARENT_COLOR;
     }
 
+  node->width = -1;
+  node->height = -1;
+  node->min_width = -1;
+  node->min_height = -1;
+  node->max_width = -1;
+  node->max_height = -1;
+
   for (i = 0; i < node->n_properties; i++)
     {
       CRDeclaration *decl = node->properties[i];
@@ -1092,7 +1134,47 @@ ensure_borders (StThemeNode *node)
         do_border_property (node, decl);
       else if (g_str_has_prefix (property_name, "padding"))
         do_padding_property (node, decl);
+      else if (strcmp (property_name, "width") == 0)
+        do_size_property (node, decl, &node->width);
+      else if (strcmp (property_name, "height") == 0)
+        do_size_property (node, decl, &node->height);
+      else if (strcmp (property_name, "min-width") == 0)
+        do_size_property (node, decl, &node->min_width);
+      else if (strcmp (property_name, "min-height") == 0)
+        do_size_property (node, decl, &node->min_height);
+      else if (strcmp (property_name, "max-width") == 0)
+        do_size_property (node, decl, &node->max_width);
+      else if (strcmp (property_name, "max-height") == 0)
+        do_size_property (node, decl, &node->max_height);
     }
+
+  if (node->width != -1)
+    {
+      if (node->min_width == -1)
+        node->min_width = node->width;
+      else if (node->width < node->min_width)
+        node->width = node->min_width;
+      if (node->max_width == -1)
+        node->max_width = node->width;
+      else if (node->width > node->max_width)
+        node->width = node->max_width;
+    }
+  else
+    node->width = node->min_width;
+
+  if (node->height != -1)
+    {
+      if (node->min_height == -1)
+        node->min_height = node->height;
+      else if (node->height < node->min_height)
+        node->height = node->min_height;
+      if (node->max_height == -1)
+        node->max_height = node->height;
+      else if (node->height > node->max_height)
+        node->height = node->max_height;
+    }
+  else
+    node->height = node->min_height;
 }
 
 double
@@ -1102,7 +1184,7 @@ st_theme_node_get_border_width (StThemeNode *node,
   g_return_val_if_fail (ST_IS_THEME_NODE (node), 0.);
   g_return_val_if_fail (side >= ST_SIDE_TOP && side <= ST_SIDE_LEFT, 0.);
 
-  ensure_borders (node);
+  ensure_geometry (node);
 
   return node->border_width[side];
 }
@@ -1114,11 +1196,65 @@ st_theme_node_get_border_radius (StThemeNode *node,
   g_return_val_if_fail (ST_IS_THEME_NODE (node), 0.);
   g_return_val_if_fail (corner >= ST_CORNER_TOPLEFT && corner <= ST_CORNER_BOTTOMLEFT, 0.);
 
-  ensure_borders (node);
+  ensure_geometry (node);
 
   return node->border_radius[corner];
 }
 
+int
+st_theme_node_get_width (StThemeNode *node)
+{
+  g_return_val_if_fail (ST_IS_THEME_NODE (node), -1);
+
+  ensure_geometry (node);
+  return node->width;
+}
+
+int
+st_theme_node_get_height (StThemeNode *node)
+{
+  g_return_val_if_fail (ST_IS_THEME_NODE (node), -1);
+
+  ensure_geometry (node);
+  return node->height;
+}
+
+int
+st_theme_node_get_min_width (StThemeNode *node)
+{
+  g_return_val_if_fail (ST_IS_THEME_NODE (node), -1);
+
+  ensure_geometry (node);
+  return node->min_width;
+}
+
+int
+st_theme_node_get_min_height (StThemeNode *node)
+{
+  g_return_val_if_fail (ST_IS_THEME_NODE (node), -1);
+
+  ensure_geometry (node);
+  return node->min_height;
+}
+
+int
+st_theme_node_get_max_width (StThemeNode *node)
+{
+  g_return_val_if_fail (ST_IS_THEME_NODE (node), -1);
+
+  ensure_geometry (node);
+  return node->max_width;
+}
+
+int
+st_theme_node_get_max_height (StThemeNode *node)
+{
+  g_return_val_if_fail (ST_IS_THEME_NODE (node), -1);
+
+  ensure_geometry (node);
+  return node->max_height;
+}
+
 static GetFromTermResult
 get_background_color_from_term (StThemeNode  *node,
                                 CRTerm       *term,
@@ -1147,6 +1283,7 @@ ensure_background (StThemeNode *node)
 
   node->background_computed = TRUE;
   node->background_color = TRANSPARENT_COLOR;
+  node->background_gradient_type = ST_GRADIENT_NONE;
 
   ensure_properties (node);
 
@@ -1242,6 +1379,35 @@ ensure_background (StThemeNode *node)
               node->background_image = NULL;
             }
         }
+      else if (strcmp (property_name, "-gradient-direction") == 0)
+        {
+          CRTerm *term = decl->value;
+          if (strcmp (term->content.str->stryng->str, "vertical") == 0)
+            {
+              node->background_gradient_type = ST_GRADIENT_VERTICAL;
+            }
+          else if (strcmp (term->content.str->stryng->str, "horizontal") == 0)
+            {
+              node->background_gradient_type = ST_GRADIENT_HORIZONTAL;
+            }
+          else if (strcmp (term->content.str->stryng->str, "radial") == 0)
+            {
+              node->background_gradient_type = ST_GRADIENT_RADIAL;
+            }
+          else
+            {
+              g_warning ("Unrecognized background-gradient-direction \"%s\"",
+                         term->content.str->stryng->str);
+            }
+        }
+      else if (strcmp (property_name, "-gradient-start") == 0)
+        {
+          get_color_from_term (node, decl->value, &node->background_color);
+        }
+      else if (strcmp (property_name, "-gradient-end") == 0)
+        {
+          get_color_from_term (node, decl->value, &node->background_gradient_end);
+        }
     }
 }
 
@@ -1304,6 +1470,34 @@ st_theme_node_get_foreground_color (StThemeNode  *node,
   *color = node->foreground_color;
 }
 
+
+/**
+ * st_theme_node_get_background_gradient:
+ * @node: A #StThemeNode
+ * @type: (out): Type of gradient
+ * @start: Color at start of gradient
+ * @end: Color at end of gradient
+ *
+ * The @start and @end arguments will only be set if @type is not #ST_GRADIENT_NONE.
+ */
+void
+st_theme_node_get_background_gradient (StThemeNode    *node,
+                                       StGradientType *type,
+                                       ClutterColor   *start,
+                                       ClutterColor   *end)
+{
+  g_return_if_fail (ST_IS_THEME_NODE (node));
+
+  ensure_background (node);
+
+  *type = node->background_gradient_type;
+  if (*type != ST_GRADIENT_NONE)
+    {
+      *start = node->background_color;
+      *end = node->background_gradient_end;
+    }
+}
+
 void
 st_theme_node_get_border_color (StThemeNode  *node,
                                 StSide        side,
@@ -1312,7 +1506,7 @@ st_theme_node_get_border_color (StThemeNode  *node,
   g_return_if_fail (ST_IS_THEME_NODE (node));
   g_return_if_fail (side >= ST_SIDE_TOP && side <= ST_SIDE_LEFT);
 
-  ensure_borders (node);
+  ensure_geometry (node);
 
   *color = node->border_color[side];
 }
@@ -1324,7 +1518,7 @@ st_theme_node_get_padding (StThemeNode *node,
   g_return_val_if_fail (ST_IS_THEME_NODE (node), 0.);
   g_return_val_if_fail (side >= ST_SIDE_TOP && side <= ST_SIDE_LEFT, 0.);
 
-  ensure_borders (node);
+  ensure_geometry (node);
 
   return node->padding[side];
 }
@@ -1805,7 +1999,10 @@ st_theme_node_get_font (StThemeNode *node)
     }
 
   if (family)
-    pango_font_description_set_family (node->font_desc, family);
+    {
+      pango_font_description_set_family (node->font_desc, family);
+      g_free (family);
+    }
 
   if (size_set)
     pango_font_description_set_absolute_size (node->font_desc, size);
@@ -1965,6 +2162,87 @@ st_theme_node_get_border_image (StThemeNode *node)
   return NULL;
 }
 
+/**
+ * st_theme_node_get_shadow:
+ * @node: a #StThemeNode
+ *
+ * Gets the value for the -st-shadow style property
+ *
+ * Return value: (transfer none): the node's shadow, or %NULL
+ *   if node has no shadow
+ */
+StShadow *
+st_theme_node_get_shadow (StThemeNode *node)
+{
+  int i;
+
+  if (node->shadow_computed)
+    return node->shadow;
+
+  node->shadow = NULL;
+  node->shadow_computed = TRUE;
+
+  ensure_properties (node);
+
+  for (i = node->n_properties - 1; i >= 0; i--)
+    {
+      CRDeclaration *decl = node->properties[i];
+
+      if (strcmp (decl->property->stryng->str, "-st-shadow") == 0)
+        {
+          /* Set value for width/color/blur in any order */
+          CRTerm *term;
+          ClutterColor color = { 0x0, 0x0, 0x0, 0xff };
+          gdouble xoffset = 0.;
+          gdouble yoffset = 0.;
+          gdouble blur = 0.;
+          int n_offsets = 0;
+
+          for (term = decl->value; term; term = term->next)
+            {
+              GetFromTermResult result;
+
+              if (term->type == TERM_NUMBER)
+                {
+                  gdouble value;
+                  gdouble multiplier;
+
+                  multiplier = (term->unary_op == MINUS_UOP) ? -1. : 1.;
+                  result = get_length_from_term (node, term, FALSE, &value);
+                  if (result != VALUE_NOT_FOUND)
+                    {
+                      switch (n_offsets++)
+                        {
+                        case 0:
+                          xoffset = multiplier * value;
+                          break;
+                        case 1:
+                          yoffset = multiplier * value;
+                          break;
+                        case 2:
+                          if (multiplier < 0)
+                              g_warning ("Negative blur values are "
+                                         "not allowed");
+                          blur = value;
+                        }
+                      continue;
+                    }
+                }
+
+              result = get_color_from_term (node, term, &color);
+              if (result != VALUE_NOT_FOUND)
+                {
+                  continue;
+                }
+            }
+          node->shadow = st_shadow_new (&color, xoffset, yoffset, blur);
+
+          return node->shadow;
+        }
+    }
+  return NULL;
+}
+
 static float
 get_width_inc (StThemeNode *node)
 {
@@ -2011,9 +2289,10 @@ st_theme_node_adjust_for_height (StThemeNode  *node,
  * @for_height: (inout): the height to adjust
  *
  * Adjusts the minimum and natural width computed for an actor by
- * adding on the necessary space for borders and padding. This is a
- * convenience function meant to be called from the get_preferred_width()
- * method of a #ClutterActor subclass
+ * adding on the necessary space for borders and padding and taking
+ * into account any minimum or maximum width. This is a convenience
+ * function meant to be called from the get_preferred_width() method
+ * of a #ClutterActor subclass
  */
 void
 st_theme_node_adjust_preferred_width (StThemeNode  *node,
@@ -2024,14 +2303,25 @@ st_theme_node_adjust_preferred_width (StThemeNode  *node,
 
   g_return_if_fail (ST_IS_THEME_NODE (node));
 
-  ensure_borders (node);
+  ensure_geometry (node);
 
   width_inc = get_width_inc (node);
 
   if (min_width_p)
-    *min_width_p += width_inc;
+    {
+      if (node->min_width != -1)
+        *min_width_p = node->min_width;
+      *min_width_p += width_inc;
+    }
+
   if (natural_width_p)
-    *natural_width_p += width_inc;
+    {
+      if (node->width != -1)
+        *natural_width_p = node->width;
+      if (node->max_width != -1)
+        *natural_width_p = MIN (*natural_width_p, node->max_width);
+      *natural_width_p += width_inc;
+    }
 }
 
 /**
@@ -2066,9 +2356,10 @@ st_theme_node_adjust_for_width (StThemeNode  *node,
  * @for_height: (inout): the height to adjust
  *
  * Adjusts the minimum and natural height computed for an actor by
- * adding on the necessary space for borders and padding. This is a
- * convenience function meant to be called from the get_preferred_height()
- * method of a #ClutterActor subclass
+ * adding on the necessary space for borders and padding and taking
+ * into account any minimum or maximum height. This is a convenience
+ * function meant to be called from the get_preferred_height() method
+ * of a #ClutterActor subclass
  */
 void
 st_theme_node_adjust_preferred_height (StThemeNode  *node,
@@ -2079,21 +2370,31 @@ st_theme_node_adjust_preferred_height (StThemeNode  *node,
 
   g_return_if_fail (ST_IS_THEME_NODE (node));
 
-  ensure_borders (node);
+  ensure_geometry (node);
 
   height_inc = get_height_inc (node);
 
   if (min_height_p)
-    *min_height_p += height_inc;
+    {
+      if (node->min_height != -1)
+        *min_height_p = node->min_height;
+      *min_height_p += height_inc;
+    }
   if (natural_height_p)
-    *natural_height_p += height_inc;
+    {
+      if (node->height != -1)
+        *natural_height_p = node->height;
+      if (node->max_height != -1)
+        *natural_height_p = MIN (*natural_height_p, node->max_height);
+      *natural_height_p += height_inc;
+    }
 }
 
 /**
  * st_theme_node_get_content_box:
  * @node: a #StThemeNode
  * @allocation: the box allocated to a #ClutterAlctor
- * @content_box: (out): computed box occupied by the actor's content
+ * @content_box: computed box occupied by the actor's content
  *
  * Gets the box within an actor's allocation that contents the content
  * of an actor (excluding borders and padding). This is a convenience function
@@ -2105,14 +2406,33 @@ st_theme_node_get_content_box (StThemeNode           *node,
                                const ClutterActorBox *allocation,
                                ClutterActorBox       *content_box)
 {
+  double noncontent_left, noncontent_top, noncontent_right, noncontent_bottom;
+  double avail_width, avail_height, content_width, content_height;
+
   g_return_if_fail (ST_IS_THEME_NODE (node));
 
-  ensure_borders (node);
+  ensure_geometry (node);
 
-  content_box->x1 = (int)(0.5 + node->border_width[ST_SIDE_LEFT]) + node->padding[ST_SIDE_LEFT];
-  content_box->y1 = (int)(0.5 + node->border_width[ST_SIDE_TOP]) + node->padding[ST_SIDE_TOP];
-  content_box->x2 = allocation->x2 - allocation->x1 - ((int)(0.5 + node->border_width[ST_SIDE_RIGHT]) + node->padding[ST_SIDE_RIGHT]);
-  content_box->y2 = allocation->y2 - allocation->y1 - ((int)(0.5 + node->border_width[ST_SIDE_BOTTOM]) + node->padding[ST_SIDE_BOTTOM]);
+  avail_width = allocation->x2 - allocation->x1;
+  avail_height = allocation->y2 - allocation->y1;
+
+  noncontent_left = node->border_width[ST_SIDE_LEFT] + node->padding[ST_SIDE_LEFT];
+  noncontent_top = node->border_width[ST_SIDE_TOP] + node->padding[ST_SIDE_TOP];
+  noncontent_right = node->border_width[ST_SIDE_RIGHT] + node->padding[ST_SIDE_RIGHT];
+  noncontent_bottom = node->border_width[ST_SIDE_BOTTOM] + node->padding[ST_SIDE_BOTTOM];
+
+  content_box->x1 = (int)(0.5 + noncontent_left);
+  content_box->y1 = (int)(0.5 + noncontent_top);
+
+  content_width = avail_width - noncontent_left - noncontent_right;
+  if (content_width < 0)
+    content_width = 0;
+  content_height = avail_height - noncontent_top - noncontent_bottom;
+  if (content_height < 0)
+    content_height = 0;
+
+  content_box->x2 = (int)(0.5 + content_box->x1 + content_width);
+  content_box->y2 = (int)(0.5 + content_box->y1 + content_height);
 }
 
 
@@ -2131,8 +2451,8 @@ st_theme_node_geometry_equal (StThemeNode *node,
 {
   StSide side;
 
-  ensure_borders (node);
-  ensure_borders (other);
+  ensure_geometry (node);
+  ensure_geometry (other);
 
   for (side = ST_SIDE_TOP; side <= ST_SIDE_LEFT; side++)
     {
@@ -2142,5 +2462,12 @@ st_theme_node_geometry_equal (StThemeNode *node,
         return FALSE;
     }
 
+  if (node->width != other->width || node->height != other->height)
+    return FALSE;
+  if (node->min_width != other->min_width || node->min_height != other->min_height)
+    return FALSE;
+  if (node->max_width != other->max_width || node->max_height != other->max_height)
+    return FALSE;
+
   return TRUE;
 }
