@@ -148,7 +148,6 @@ st_tooltip_get_preferred_width (ClutterActor *self,
 {
   StTooltipPrivate *priv = ST_TOOLTIP (self)->priv;
   StThemeNode *theme_node = st_widget_get_theme_node (ST_WIDGET (self));
-  gfloat min_label_w, natural_label_w;
   gfloat label_height, arrow_height;
   ClutterActor *arrow_image;
 
@@ -180,13 +179,8 @@ st_tooltip_get_preferred_width (ClutterActor *self,
     {
       clutter_actor_get_preferred_width (priv->label,
                                          label_height,
-                                         &min_label_w,
-                                         &natural_label_w);
-    }
-  else
-    {
-      min_label_w = 0;
-      natural_label_w = 0;
+                                         min_width_p,
+                                         natural_width_p);
     }
 
   st_theme_node_adjust_preferred_width (theme_node, min_width_p, natural_width_p);
@@ -355,6 +349,20 @@ st_tooltip_unmap (ClutterActor *self)
   clutter_actor_unmap (priv->label);
 }
 
+static void
+st_tooltip_dispose (GObject *self)
+{
+  StTooltipPrivate *priv = ST_TOOLTIP (self)->priv;
+
+  if (priv->label)
+    {
+      clutter_actor_destroy (priv->label);
+      priv->label = NULL;
+    }
+
+  G_OBJECT_CLASS (st_tooltip_parent_class)->dispose (self);
+}
+
 static void
 st_tooltip_class_init (StTooltipClass *klass)
 {
@@ -367,6 +375,7 @@ st_tooltip_class_init (StTooltipClass *klass)
 
   gobject_class->set_property = st_tooltip_set_property;
   gobject_class->get_property = st_tooltip_get_property;
+  gobject_class->dispose = st_tooltip_dispose;
 
   actor_class->get_preferred_width = st_tooltip_get_preferred_width;
   actor_class->get_preferred_height = st_tooltip_get_preferred_height;
